from typing import Callable, Union, Tuple, Literal, Optional
from contextlib import contextmanager


class RemoveListener(Exception):
    pass


class GameEventHandler:
    """
    This class is responsible for notifying any registered listeners when a game event is triggered.
    Game events can be anything related to a change in the game's state (for example - changing the current room,
    a character dying, a spell being cast etc.) and are triggered by calling .on_event()
    with the relevant event_key and any other related information for that event.

    Listeners are categorised by event_key, which is either a string representing the type of event occurring, or
    a tuple containing two strings (the first being either "before" or "after" and the second representing event type).
    A tuple event key such as this will be generated when using this class as a contextmanager;
    in such a case, .on_event() is invoked both before and after the event logic contained in the `with` block, with
    the appropriate tuple being passed as an event key each time.

    Alternatively, event_key can be omitted when adding (or removing) a listener, to indicate that the listener
    should be notified on all events
    """

    def __init__(self):
        self._listeners = {}

    @contextmanager
    def __call__(self, event_type: str, *args, **kwargs):
        self.on_event(("before", event_type), *args, **kwargs)
        yield
        self.on_event(("after", event_type), *args, **kwargs)

    def add_listener(
            self, listener: Callable,
            event_key: Optional[Union[str, Tuple[Literal["before", "after"], str]]] = None
    ) -> None:
        self._listeners.setdefault(event_key, set()).add(listener)

    def remove_listener(
            self, listener: Callable,
            event_key: Optional[Union[str, Tuple[Literal["before", "after"], str]]] = None
    ) -> None:
        self._listeners.setdefault(event_key, set()).remove(listener)

    def on_event(self, event_key: Union[str, Tuple[Literal["before", "after"], str]], *args, **kwargs) -> None:
        # Notify listeners registered to all events
        listeners_to_remove = set()
        for listener in self._listeners.get(None, set()):
            try:
                listener(None, *args, **kwargs)
            except RemoveListener:
                listeners_to_remove.add(listener)

        for listener in listeners_to_remove:
            self.remove_listener(listener, None)

        # Notify listeners registered specifically to this event
        listeners_to_remove = set()
        for listener in self._listeners.get(event_key, set()):
            try:
                listener(event_key, *args, **kwargs)
            except RemoveListener:
                listeners_to_remove.add(listener)

        for listener in listeners_to_remove:
            self.remove_listener(listener, event_key)



from sys import modules
from typing import Dict


class CustomClassHandler:
    def __init__(self, game):
        self._game = game

        self._classes = {**self._game.config.CUSTOM_CLASSES}

    def register(self, value: Dict[str, type]) -> None:
        self._classes.update(value)

    def get(self, class_name: str) -> type:
        """
        Attempts to locate a class using the provided class name.
        First checks locally stored classes (initially populated from Config.CUSTOM_CLASSES
        but additional classes can be manually added during runtime),
        then checks classes in the global namespace (in __main__).
        Supports dot notation to access objects stored under module or class attributes in the global namespace
        """

        if class_name in self._classes:
            return self._classes[class_name]

        try:
            nodes = class_name.split(".")

            result = getattr(modules["__main__"], nodes.pop(0))
            for node in nodes:
                result = getattr(result, node)

            return result

        except AttributeError:
            raise ValueError(f"unable to resolve the class name '{class_name}'")



from typing import Type, Tuple, List
from os import path

from ..methods import Methods
from .renderable import Renderable
from .entity import Entity
from .enums import EntityDataKey


class Room(Renderable):
    """
    Concrete class tightly coupled to the World screen, which renders a single room and its occupants.
    Can be subclassed as needed to add further functionality
    """

    def __init__(self, parent: "World", room_id: str):
        super().__init__(parent.game, parent=parent, render_position=(0, 0), priority=0)

        self._room_id = room_id
        self.surface = self._generate_surface()

        self._is_loaded = False

    @property
    def room_id(self) -> str:
        return self._room_id

    def _update(self, elapsed_ms: int, input_events: list, *args, **kwargs) -> None:
        if not self._is_loaded:
            self._load_room()

            self._is_loaded = True

    def _generate_surface(self):
        """
        Loads the background image for the room object as a new Surface.
        Assumes a standard location for the image file as dictated below in `background_file_path`
        """

        background_file_path = path.join(
            self.game.config.RESOURCE_FOLDER_PATH,
            self.game.screen.state.registered_get("room_background_file_path", [self._room_id])
        )

        return Methods.load_image(background_file_path).convert_alpha()

    def _load_room(self):
        """
        Instantiates all the entities present in the current room.
        Assumes that any custom Entity subclasses listed in the game's state
        have been made available to the game's CustomClassHandler
        """

        curr_room_entities_ids: List[str] = self.game.screen.state.registered_get(
            "room_entities_ids", [self._room_id]
        )

        for entity_id in curr_room_entities_ids:
            entity_data: dict = self.game.screen.state.registered_get("entity", [entity_id])

            entity_class: Type[Entity] = self.game.custom_class_handler.get(entity_data[EntityDataKey.CLASS])
            entity_details: Tuple[list, dict] = (
                entity_data.get(EntityDataKey.ARGS, []),
                entity_data.get(EntityDataKey.KWARGS, {})
            )

            entity_class(self, *entity_details[0], **entity_details[1])



from objectextensions import Decorators

from typing import Dict, Type, Callable

from .hitbox import Hitbox


class RecurfaceHitbox(Hitbox):
    @Decorators.classproperty
    def COLLISION_CHECKERS(cls) -> Dict[Type[Hitbox], Callable[[Hitbox, Hitbox], bool]]:
        return {
            cls: cls._is_collision_recurfacehitbox
        }

    @staticmethod
    def _is_collision_recurfacehitbox(a: "RecurfaceHitbox", b: "RecurfaceHitbox") -> bool:
        a_x_bounds = (
            a.parent_renderable.x_render_position,
            a.parent_renderable.x_render_position + a.parent_renderable.surface.get_width()
        )
        a_y_bounds = (
            a.parent_renderable.y_render_position,
            a.parent_renderable.y_render_position + a.parent_renderable.surface.get_height()
        )

        b_x_bounds = (
            b.parent_renderable.x_render_position,
            b.parent_renderable.x_render_position + b.parent_renderable.surface.get_width()
        )
        b_y_bounds = (
            b.parent_renderable.y_render_position,
            b.parent_renderable.y_render_position + b.parent_renderable.surface.get_height()
        )

        return (
                (a_x_bounds[0] < b_x_bounds[1]) and (b_x_bounds[0] < a_x_bounds[1]) and
                (a_y_bounds[0] < b_y_bounds[1]) and (b_y_bounds[0] < a_y_bounds[1])
        )
